from mpi4py import MPI

import jax
import jax.numpy as jnp

import numpy as np

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_allreduce():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, token = allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_jit():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_scalar():
    from mpi4jax import allreduce

    arr = 1
    _arr = 1

    res, token = allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_scalar_jit():
    from mpi4jax import allreduce

    arr = 1
    _arr = 1

    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_vmap():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res = jax.vmap(lambda x: allreduce(x, op=MPI.SUM)[0], in_axes=0, out_axes=0)(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_vmap_jit():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res = jax.jit(
        jax.vmap(lambda x: allreduce(x, op=MPI.SUM)[0], in_axes=0, out_axes=0)
    )(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_transpose():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(_arr)
    assert jnp.array_equal(_arr, res)


def test_allreduce_transpose_jit():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    def f(x):
        (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(x)
        return res

    res = jax.jit(f)(arr)
    assert jnp.array_equal(_arr, res)


def test_allreduce_transpose2():
    # test transposing twice
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()
    _arr2 = arr.copy()

    def lt(y):
        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(y)[0]

    (res,) = jax.linear_transpose(lt, _arr)(_arr2)
    expected, _ = allreduce(_arr2, op=MPI.SUM)
    assert jnp.array_equal(expected, res)


def test_allreduce_transpose2_jit():
    # test transposing twice
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()
    _arr2 = arr.copy()

    def lt(y):
        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(y)[0]

    def f(x):
        (res,) = jax.linear_transpose(lt, _arr)(x)
        return res

    res = jax.jit(f)(_arr2)
    expected, _ = allreduce(_arr2, op=MPI.SUM)
    assert jnp.array_equal(expected, res)


def test_allreduce_grad():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, grad = jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM)[0].sum())(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)

    res, grad = jax.jit(
        jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM)[0].sum())
    )(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)

    def testfun(x):
        y, token = allreduce(x, op=MPI.SUM)
        z = x + 2 * y  # noqa: F841
        res, token = allreduce(x, op=MPI.SUM, token=token)
        return res.sum()

    res, grad = jax.jit(jax.value_and_grad(testfun))(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)


def test_allreduce_jvp():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, jvp = jax.jvp(lambda x: allreduce(x, op=MPI.SUM)[0], (arr,), (_arr,))

    expected, _ = allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(expected, res)
    expected, _ = allreduce(_arr, op=MPI.SUM)
    assert jnp.array_equal(expected, jvp)


def test_allreduce_vjp():
    from mpi4jax import allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, vjp_fun = jax.vjp(lambda x: allreduce(x, op=MPI.SUM)[0], arr)
    (vjp,) = vjp_fun(_arr)

    expected, _ = allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(expected, res)
    assert jnp.array_equal(_arr, vjp)


def test_allreduce_chained():
    from mpi4jax import allreduce

    def foo(x):
        token = jax.lax.create_token()
        x1, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
        x2, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
        return x1 + x2

    res_t = jax.grad(foo)(0.0)

    expected = 2.0
    assert jnp.array_equal(expected, res_t)


def test_allreduce_chained_jit():
    from mpi4jax import allreduce

    def foo(x):
        token = jax.lax.create_token()
        x1, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
        x2, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
        return x1 + x2

    res_t = jax.jit(jax.grad(foo))(0.0)

    expected = 2.0
    np.testing.assert_allclose(expected, res_t)


def test_custom_vjp():
    from mpi4jax import allreduce

    # define an arbitrary function with custom_vjp
    @jax.custom_vjp
    def f(x, y):
        r = jnp.sin(x) * y
        r = r.sum()
        return allreduce(r, op=MPI.SUM)[0]

    def f_fwd(x, y):
        # Returns primal output and residuals to be used in backward pass by f_bwd.
        return f(x, y), (jnp.cos(x), jnp.sin(x), y)

    def f_bwd(res, g):
        g = allreduce(g, op=MPI.SUM)[0]
        cos_x, sin_x, y = res  # Gets residuals computed in f_fwd
        return (cos_x * g * y, sin_x * g)

    f.defvjp(f_fwd, f_bwd)

    # check that it does not crash
    _ = jax.jit(f)(jnp.ones(3), jnp.ones(3) * 2)
    _ = jax.jit(jax.grad(f))(jnp.ones(3), jnp.ones(3) * 2)


def test_advanced_jvp():
    """
    This is an integration test checking that the effects introduced by mpi4jax
    are correctly handled by jax in a complex `custom_vjp` rule computing a
    `jax.grad` inside of the rule itself.

    This code essentially computes the mathematically correct gradient of an
    expectation value of a function over a probability distribution.
    It is taken from netket.jax.expect
    https://github.com/netket/netket/blob/master/netket/jax/_expect.py
    """
    from mpi4jax import allreduce
    from functools import partial

    def expect(
        log_pdf,
        expected_fun,
        pars,
        x,
        *expected_fun_args,
        n_chains,
    ):
        return _expect(n_chains, log_pdf, expected_fun, pars, x, *expected_fun_args)

    @partial(jax.custom_vjp, nondiff_argnums=(0, 1, 2))
    def _expect(n_chains, log_pdf, expected_fun, pars, x, *expected_fun_args):
        L_x = expected_fun(pars, x, *expected_fun_args).reshape((n_chains, -1))
        return allreduce(L_x.mean(), op=MPI.SUM)[0] / MPI.COMM_WORLD.Get_size()

    def _expect_fwd(n_chains, log_pdf, expected_fun, pars, x, *expected_fun_args):
        L_x = expected_fun(pars, x, *expected_fun_args)
        L_x_r = L_x.reshape((n_chains, -1))
        dL_x = allreduce(L_x_r.mean(), op=MPI.SUM)[0] / MPI.COMM_WORLD.Get_size()
        ΔL_x = L_x - dL_x
        return dL_x, (pars, x, expected_fun_args, ΔL_x)

    def _expect_bwd(n_chains, log_pdf, expected_fun, residuals, dout):
        pars, x, cost_args, ΔL_x = residuals
        dL = dout

        def f(pars, x, *cost_args):
            log_p = log_pdf(pars, x)
            term1 = jax.vmap(jnp.multiply)(ΔL_x, log_p)
            term2 = expected_fun(pars, x, *cost_args)
            out = (
                allreduce(jnp.mean(term1 + term2, axis=0), op=MPI.SUM)[0]
                / MPI.COMM_WORLD.Get_size()
            )
            out = out.sum()
            return out

        aa, pb = jax.vjp(f, pars, x, *cost_args)
        grad_f = pb(dL)
        return grad_f

    _expect.defvjp(_expect_fwd, _expect_bwd)

    def log_pdf(w, x):
        return jnp.sum(x.dot(w), axis=-1)

    def expected_fun(w, x, *args):
        return jnp.exp(jnp.sum(x.dot(w), axis=-1)) - 2

    w = jax.random.normal(jax.random.PRNGKey(3), (4, 8))
    x = jax.random.normal(jax.random.PRNGKey(3), (16, 4))

    expect(log_pdf, expected_fun, w, x, None, n_chains=16)
    O, vjpfun = jax.vjp(
        lambda w: expect(log_pdf, expected_fun, w, x, None, n_chains=16), w
    )
    vjpfun(jnp.ones_like(O))



import jax
from jax import linear_util as lu
from jax.interpreters import xla

# registry for wrapped mpi4jax ops
from .register_overrides import token_override_registry  # noqa: E402

# registry for wrapped JAX primitives
recursive_token_forwarding_registry = {}


def safe_map(f, *args):
    args = [list(arg) for arg in args]
    for arg in args:
        assert len(arg) == len(args[0]), f"Mismatched lengths: {[len(x) for x in args]}"
    return [f(*arg) for arg in zip(*args)]


def xla_call_overrride(read, eqn, token):
    subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
    bind_params["donated_invars"] = (False,) + bind_params["donated_invars"]
    map_token = lambda func: lu.wrap_init(  # noqa: E731
        lambda token, *args: _token_forwarding(func.call_wrapped, token)(*args)
    )
    subfuns = safe_map(map_token, subfuns)
    ans = eqn.primitive.bind(
        *subfuns, token, *safe_map(read, eqn.invars), **bind_params
    )
    token = ans[0]
    ans = ans[1:]  # Drop the token.
    return token, ans


recursive_token_forwarding_registry[xla.xla_call_p] = xla_call_overrride


def scan_override(read, eqn, token):
    _, bind_params = eqn.primitive.get_bind_params(eqn.params)
    new_body_fn = lambda token, *args: _token_forwarding(  # noqa: E731
        jax.core.jaxpr_as_fun(bind_params["jaxpr"]), token
    )(*args)
    bind_params["jaxpr"] = jax.make_jaxpr(new_body_fn)(
        token, *safe_map(read, eqn.invars)
    )
    # Update bind_params to account for the additional token.
    bind_params["num_carry"] += 1
    bind_params["linear"] = (False,) + bind_params["linear"]
    ans = eqn.primitive.bind(token, *safe_map(read, eqn.invars), **bind_params)
    token = ans[0]
    ans = ans[1:]  # Drop the token.
    return token, ans


recursive_token_forwarding_registry[jax.lax.scan_p] = scan_override


def while_override(read, eqn, token):
    _, bind_params = eqn.primitive.get_bind_params(eqn.params)
    new_body_fn = lambda token, *args: _token_forwarding(  # noqa: E731
        jax.core.jaxpr_as_fun(bind_params["body_jaxpr"]), token
    )(*args)
    bind_params["body_jaxpr"] = jax.make_jaxpr(new_body_fn)(
        token, *safe_map(read, eqn.invars)
    )
    # We use `auto_tokenize` here since the condition function
    # is forced to only return a boolean value and cannot return
    # a token.
    new_cond_fn = lambda token, *args: _token_forwarding(  # noqa: E731
        jax.core.jaxpr_as_fun(bind_params["cond_jaxpr"]), token
    )(*args)[1:]
    bind_params["cond_jaxpr"] = jax.make_jaxpr(new_cond_fn)(
        token, *safe_map(read, eqn.invars)
    )
    # Update bind_params to account for the additional token.
    ans = eqn.primitive.bind(token, *safe_map(read, eqn.invars), **bind_params)
    token = ans[0]
    ans = ans[1:]  # Drop the token.
    return token, ans


recursive_token_forwarding_registry[jax.lax.while_p] = while_override


def cond_override(read, eqn, token):
    _, bind_params = eqn.primitive.get_bind_params(eqn.params)
    branch_jaxprs = []
    cond_var = eqn.invars[0]
    other_vars = eqn.invars[1:]
    for branch in bind_params["branches"]:
        new_branch_fn = lambda token, *args: _token_forwarding(  # noqa: E731
            jax.core.jaxpr_as_fun(branch), token
        )(*args)
        new_jaxpr = jax.make_jaxpr(new_branch_fn)(token, *safe_map(read, other_vars))
        branch_jaxprs.append(new_jaxpr)
    bind_params["branches"] = tuple(branch_jaxprs)
    bind_params["linear"] = (False,) + bind_params["linear"]
    ans = eqn.primitive.bind(
        read(cond_var), token, *safe_map(read, other_vars), **bind_params
    )
    token = ans[0]
    ans = ans[1:]  # Drop the token.
    return token, ans


recursive_token_forwarding_registry[jax.lax.cond_p] = cond_override


def _override_tokens(jaxpr, consts, token, *args):
    if token is None:  # Create a new token if one is not passed.
        token = jax.lax.create_token()

    def read(v):
        if type(v) is jax.core.Literal:
            return v.val
        else:
            return env[v]

    def write(v, val):
        env[v] = val

    env = {}

    # compatibility with jax<0.3.10
    if hasattr(jax.core, "unitvar"):
        write(jax.core.unitvar, jax.core.unit)

    safe_map(write, jaxpr.constvars, consts)
    safe_map(write, jaxpr.invars, args)
    for eqn in jaxpr.eqns:
        # Here, we override the original primatives with our own
        # token forcing method.
        if eqn.primitive in token_override_registry:
            token_override = token_override_registry[eqn.primitive]
            ans = token_override(
                safe_map(read, eqn.invars), new_token=token, **eqn.params
            )
            # We pass along the new token returned by the above binding
            token = ans[-1]
            safe_map(write, eqn.outvars, ans)
        else:
            if eqn.primitive in recursive_token_forwarding_registry:
                rewrite = recursive_token_forwarding_registry[eqn.primitive]
                token, ans = rewrite(read, eqn, token)
            # Here, we are just reapplying the original operation if
            # not part of our communication protocol.
            # This code is mostly taken from jax.core.eval_jaxpr
            else:
                subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
                ans = eqn.primitive.bind(
                    *subfuns, *safe_map(read, eqn.invars), **bind_params
                )
            if eqn.primitive.multiple_results:
                safe_map(write, eqn.outvars, ans)
            else:
                write(eqn.outvars[0], ans)
    return (token,) + tuple(safe_map(read, jaxpr.outvars))


def _token_forwarding(f, token):
    def wrapper(*args, **kwargs):
        jaxpr = jax.make_jaxpr(f)(*args, **kwargs)
        return _override_tokens(jaxpr.jaxpr, jaxpr.consts, token, *args, **kwargs)

    return wrapper


def auto_tokenize(f):
    """Automatically manage tokens between all mpi4jax ops.

    Supports most JAX operations, including ``fori_loop``, ``cond``, ``jit``,
    ``while_loop``, and ``scan``. This transforms *all* operations in the decorated
    function, even through subfunctions and nested applications of ``jax.jit``.

    .. note::

       This transforms overrides all mpi4jax token management completely.
       Do not use this transform if you need manual control over the token managment.

    Arguments:
        f: Any function that uses mpi4jax primitives (jitted or not).

    Returns:
        A transformed version of ``f`` that automatically manages all mpi4jax tokens.

    Example:

        >>> @auto_tokenize
        ... def f(a):
        ...     # no token handling necessary
        ...     res, _ = allreduce(a, op=MPI.SUM)
        ...     res, _ = allreduce(res, op=MPI.SUM)
        ...     return res
        >>> arr = jnp.ones((3, 2))
        >>> res = f(arr)

    """

    def wrapper(*args, **kwargs):
        jaxpr, pytree = jax.make_jaxpr(f, return_shape=True)(*args, **kwargs)
        _, pytree = jax.tree_util.tree_flatten(pytree)
        res = _override_tokens(jaxpr.jaxpr, jaxpr.consts, None, *args, **kwargs)
        return jax.tree_util.tree_unflatten(pytree, res[1:])  # Drop the token.

    return wrapper



import jax
from mpi4jax._src.jax_compat import versiontuple

if versiontuple(jax.__version__) >= versiontuple("0.4.4"):
    raise ImportError(
        "mpi4jax.experimental is not yet compatible with JAX versions >= 0.4.4. "
        "Please use jax<=0.4.3 for now."
    )

from .tokenizer import auto_tokenize

__all__ = [
    "auto_tokenize",
]



import pytest

from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def run_in_subprocess(code, test_file, timeout=10):
    """Runs given code string in a subprocess"""
    import os
    import subprocess
    import sys
    from textwrap import dedent

    # this enables us to measure coverage in subprocesses
    cov_preamble = dedent(
        """
    try:
        import coverage
        coverage.process_startup()
    except ImportError:
        pass

    """
    )

    test_file.write_text(cov_preamble + code)

    # passing a mostly empty env seems to be the only way to
    # force MPI to initialize again
    env = dict(
        PATH=os.getenv("PATH", ""),
        PYTHONPATH=os.getenv("PYTHONPATH", ""),
        COVERAGE_PROCESS_START="pyproject.toml",
        XLA_PYTHON_CLIENT_PREALLOCATE="false",
        LD_LIBRARY_PATH=os.getenv("LD_LIBRARY_PATH", ""),
    )

    # non-standard Intel MPI env var for libfabric.
    if "FI_PROVIDER_PATH" in os.environ:
        env["FI_PROVIDER_PATH"] = os.getenv("FI_PROVIDER_PATH")

    proc = subprocess.run(
        [sys.executable, test_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
        timeout=timeout,
        universal_newlines=True,
        env=env,
    )
    return proc


@pytest.mark.skipif(rank > 0, reason="Runs only on rank 0")
def test_abort_on_error(tmp_path):
    from textwrap import dedent

    # test in subprocess so we don't kill the testing process itself
    test_script = dedent(
        """
        import jax
        import jax.numpy as jnp

        from mpi4py import MPI
        from mpi4jax import send

        comm = MPI.COMM_WORLD
        assert comm.Get_size() == 1

        # send to non-existing rank
        @jax.jit
        def send_jit(x):
            send(x, dest=100, comm=comm)

        send_jit(jnp.ones(10))
    """
    )

    proc = run_in_subprocess(test_script, tmp_path / "abort.py")
    print(proc.stderr)
    assert proc.returncode != 0
    assert "r0 | MPI_Send returned error code" in proc.stderr


@pytest.mark.skipif(rank > 0, reason="Runs only on rank 0")
def test_deadlock_on_exit(tmp_path):
    from textwrap import dedent

    # without our custom atexit handler, this would deadlock most of the time
    test_script = dedent(
        """
        import jax
        import jax.numpy as jnp

        from mpi4py import MPI
        from mpi4jax import sendrecv

        comm = MPI.COMM_WORLD
        assert comm.Get_size() == 1

        # sendrecv to self
        jax.jit(
            lambda x: sendrecv(sendbuf=x, recvbuf=x, source=0, dest=0, comm=comm)[0]
        )(jnp.ones(10))
    """
    )

    proc = run_in_subprocess(test_script, tmp_path / "deadlock_on_exit.py")
    assert proc.returncode == 0, proc.stderr


def test_debug_logging(capsys):
    import re
    from mpi4jax import allreduce
    from mpi4jax._src.xla_bridge.mpi_xla_bridge import set_logging

    arr = jnp.ones((3, 2))

    set_logging(True)
    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
    res.block_until_ready()

    captured = capsys.readouterr().out
    start_msg, end_msg, _ = captured.split("\n")
    assert re.match(
        rf"r{rank} \| \w{{8}} \| MPI_Allreduce with {arr.size} items", start_msg
    )
    assert re.match(
        rf"r{rank} \| \w{{8}} \| MPI_Allreduce done with code 0 \(\d\.\d{{2}}e[+-]?\d+s\)",
        end_msg,
    )

    set_logging(False)
    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
    res.block_until_ready()

    captured = capsys.readouterr().out
    assert not captured


def test_set_logging_from_envvar():
    import os
    import importlib

    from mpi4jax._src import xla_bridge
    from mpi4jax._src.xla_bridge.mpi_xla_bridge import set_logging, get_logging

    os.environ["MPI4JAX_DEBUG"] = "1"
    importlib.reload(xla_bridge)
    assert get_logging()

    os.environ["MPI4JAX_DEBUG"] = "0"
    importlib.reload(xla_bridge)
    assert not get_logging()

    set_logging(True)
    assert get_logging()

    del os.environ["MPI4JAX_DEBUG"]



import sys
import string
import random
from time import perf_counter

from libc.stdint cimport uintptr_t

from mpi4py.libmpi cimport (
    MPI_Abort,
    MPI_Comm,
    MPI_Comm_rank,
    MPI_Datatype,
    MPI_Error_string,
    MPI_MAX_ERROR_STRING,
    MPI_Op,
    MPI_Status,
    MPI_STATUS_IGNORE,
    MPI_SUCCESS,
)

cimport mpi4py.libmpi as libmpi

# MPI_STATUS_IGNORE is not exposed to Python by mpi4py, so we
# export its memory address as Python int here.
# This can then be passed to all functions that expect
# MPI_Status* instead of a pointer to a real status object.
MPI_STATUS_IGNORE_ADDR = int(<uintptr_t>MPI_STATUS_IGNORE)


#
# Logging
#


cdef bint PRINT_DEBUG = False


cpdef void set_logging(bint enable):
    global PRINT_DEBUG
    PRINT_DEBUG = enable


cpdef bint get_logging():
    return PRINT_DEBUG


cdef inline void print_debug(unicode message, unicode rid, MPI_Comm comm) nogil:
    cdef int rank
    MPI_Comm_rank(comm, &rank)
    with gil:
        print(f"r{rank} | {rid} | {message}", flush=True)


cdef unicode random_id():
    cdef unicode alphabet = (
        string.ascii_lowercase
        + string.ascii_uppercase
        + string.digits
    )
    return ''.join(random.choices(alphabet, k=8))


#
# Error handling
#

cdef inline int abort(int ierr, MPI_Comm comm, unicode message) nogil:
    with gil:
        sys.stderr.write(message)
        sys.stderr.flush()

    return MPI_Abort(comm, ierr)


cdef inline int abort_on_error(int ierr, MPI_Comm comm, unicode mpi_op) nogil:
    if ierr == MPI_SUCCESS:
        return 0

    cdef int rank
    MPI_Comm_rank(comm, &rank)

    cdef char c_string[MPI_MAX_ERROR_STRING]
    cdef int length = 0

    MPI_Error_string(ierr, c_string, &length)

    with gil:
        strerr = c_string[:length]
        message = f"r{rank} | MPI_{mpi_op} returned error code {ierr}: {strerr} - aborting\n"

    return abort(ierr, comm, message)


#
# Wrapped MPI primitives
#
cdef void mpi_barrier(MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(f"MPI_Barrier", rid, comm)
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Barrier(comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(f"MPI_Barrier done with code {ierr} ({end - start:.2e}s)", rid, comm)

    abort_on_error(ierr, comm, u"Barrier")


cdef void mpi_allgather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                        void* recvbuf, int recvcount, MPI_Datatype recvtype,
                        MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Allgather sending {sendcount}, receiving {recvcount} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Allgather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, comm
    )

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Allgather done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Allgather")


cdef void mpi_allreduce(void* sendbuf, void * recvbuf, int nitems,
                        MPI_Datatype dtype, MPI_Op op, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(f"MPI_Allreduce with {nitems} items", rid, comm)
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Allreduce(sendbuf, recvbuf, nitems, dtype, op, comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(f"MPI_Allreduce done with code {ierr} ({end - start:.2e}s)", rid, comm)

    abort_on_error(ierr, comm, u"Allreduce")


cdef void mpi_alltoall(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                       void* recvbuf, int recvcount, MPI_Datatype recvtype,
                       MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Alltoall sending {sendcount}, receiving {recvcount} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Alltoall(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, comm
    )

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Alltoall done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Alltoall")


cdef void mpi_bcast(void* sendrecvbuf, int nitems, MPI_Datatype dtype,
                    int root, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Bcast -> {root} with {nitems} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Bcast(sendrecvbuf, nitems, dtype, root, comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Bcast done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Bcast")


cdef void mpi_gather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                     void* recvbuf, int recvcount, MPI_Datatype recvtype,
                     int root, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Gather -> {root} sending {sendcount}, "
                f"receiving {recvcount} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Gather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm
    )

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Gather done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Gather")


cdef void mpi_recv(void* recvbuf, int nitems, MPI_Datatype dtype, int source,
                   int tag, MPI_Comm comm, MPI_Status* status) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Recv <- {source} with tag {tag} and {nitems} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Recv(recvbuf, nitems, dtype, source, tag, comm, status)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Recv done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Recv")


cdef void mpi_reduce(void* sendbuf, void* recvbuf, int nitems,
                     MPI_Datatype dtype, MPI_Op op, int root,
                     MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Reduce -> {root} with {nitems} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Reduce(sendbuf, recvbuf, nitems, dtype, op, root, comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Reduce done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Reduce")


cdef void mpi_scan(void* sendbuf, void* recvbuf, int nitems,
                   MPI_Datatype dtype, MPI_Op op, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Scan with {nitems} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Scan(sendbuf, recvbuf, nitems, dtype, op, comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Scan done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Scan")


cdef void mpi_scatter(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                      void* recvbuf, int recvcount, MPI_Datatype recvtype,
                      int root, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Scatter -> {root} sending {sendcount}, "
                f"receiving {recvcount} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Scatter(
        sendbuf, sendcount, sendtype,
        recvbuf, recvcount, recvtype,
        root, comm
    )

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Scatter done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Scatter")


cdef void mpi_send(void* sendbuf, int nitems, MPI_Datatype dtype,
                   int destination, int tag, MPI_Comm comm) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Send -> {destination} with tag {tag} and {nitems} items",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Send(sendbuf, nitems, dtype, destination, tag, comm)

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Send done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Send")


cdef void mpi_sendrecv(
    void* sendbuf, int sendcount, MPI_Datatype sendtype, int dest, int sendtag,
    void* recvbuf, int recvcount, MPI_Datatype recvtype, int source, int recvtag,
    MPI_Comm comm, MPI_Status* status
) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            rid = random_id()
            print_debug(
                f"MPI_Sendrecv <- {source} (tag {recvtag}, {recvcount} items) / "
                f"-> {dest} (tag {sendtag}, {sendcount} items) ",
                rid,
                comm
            )
            start = perf_counter()

    # MPI Call
    ierr = libmpi.MPI_Sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status
    )

    if PRINT_DEBUG:
        with gil:
            end = perf_counter()
            print_debug(
                f"MPI_Sendrecv done with code {ierr} ({end - start:.2e}s)",
                rid,
                comm
            )

    abort_on_error(ierr, comm, u"Sendrecv")



from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_allgather():
    from mpi4jax import allgather

    arr = jnp.ones((3, 2)) * rank

    res, _ = allgather(arr)
    for p in range(size):
        assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)


def test_allgather_jit():
    from mpi4jax import allgather

    arr = jnp.ones((3, 2)) * rank

    res = jax.jit(lambda x: allgather(x)[0])(arr)
    for p in range(size):
        assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)


def test_allgather_scalar():
    from mpi4jax import allgather

    arr = rank
    res, _ = allgather(arr)
    assert jnp.array_equal(res, jnp.arange(size))


def test_allgather_scalar_jit():
    from mpi4jax import allgather

    arr = rank
    res = jax.jit(lambda x: allgather(x)[0])(arr)
    assert jnp.array_equal(res, jnp.arange(size))



import os
import warnings
import functools

# global variables to keep track of state
_cuda_mpi_setup_done = False


def ensure_gpu_ext():
    from .xla_bridge import HAS_GPU_EXT

    if not HAS_GPU_EXT:
        raise ImportError(
            "The mpi4jax GPU extensions could not be imported. "
            "Please re-build mpi4jax with CUDA support and try again."
        )


def _is_truthy(str_val):
    return str_val.lower() in ("true", "1", "on")


def _is_falsy(str_val):
    return str_val.lower() in ("false", "0", "off")


def setup_cuda_mpi():
    global _cuda_mpi_setup_done

    if _cuda_mpi_setup_done:
        return

    _cuda_mpi_setup_done = True

    gpu_copy_behavior = os.getenv("MPI4JAX_USE_CUDA_MPI", "")

    if _is_truthy(gpu_copy_behavior):
        has_cuda_mpi = True
    elif _is_falsy(gpu_copy_behavior):
        has_cuda_mpi = False
    else:
        has_cuda_mpi = False
        warn_msg = (
            "Not using CUDA-enabled MPI. "
            "If you are sure that your MPI library is built with CUDA support, "
            "set MPI4JAX_USE_CUDA_MPI=1. To silence this warning, "
            "set MPI4JAX_USE_CUDA_MPI=0."
        )
        warnings.warn(warn_msg)

    from .xla_bridge import mpi_xla_bridge_gpu

    mpi_xla_bridge_gpu.set_copy_to_host(not has_cuda_mpi)


def translation_rule_cpu(func):
    """XLA primitive translation rule on CPU for mpi4jax custom calls.

    This runs generic setup and boilerplate functions.
    """
    # NOTE: currently does nothing, but we keep it for consistency

    # functions to call before running the translation rule
    setup_funcs = ()

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        for f in setup_funcs:
            f()
        return func(*args, **kwargs)

    return wrapped


def translation_rule_gpu(func):
    """XLA primitive translation rule on GPU for mpi4jax custom calls.

    This runs generic setup and boilerplate functions.
    """
    # functions to call before running the translation rule
    setup_funcs = (
        ensure_gpu_ext,
        setup_cuda_mpi,
    )

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        for f in setup_funcs:
            f()
        return func(*args, **kwargs)

    return wrapped



import pytest
from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_scatter():
    from mpi4jax import scatter

    if rank == 0:
        arr = jnp.stack([jnp.ones((3, 2)) * r for r in range(size)], axis=0)
    else:
        arr = jnp.ones((3, 2)) * rank

    res, _ = scatter(arr, root=0)
    assert jnp.array_equal(res, jnp.ones((3, 2)) * rank)


def test_scatter_jit():
    from mpi4jax import scatter

    if rank == 0:
        arr = jnp.stack([jnp.ones((3, 2)) * r for r in range(size)], axis=0)
    else:
        arr = jnp.ones((3, 2)) * rank

    res = jax.jit(lambda x: scatter(x, root=0)[0])(arr)
    assert jnp.array_equal(res, jnp.ones((3, 2)) * rank)


@pytest.mark.skipif(rank > 0, reason="Only runs on root process")
def test_scatter_wrong_size():
    from mpi4jax import scatter

    arr = jnp.ones((size + 1,)) * rank
    with pytest.raises(ValueError) as excinfo:
        scatter(arr, root=0)

    assert "Scatter input must have shape" in str(excinfo.value)



import functools

from mpi4py import MPI as _MPI

import numpy as _np

from jax.interpreters import xla, mlir
import jaxlib.mlir.ir as ir
from jaxlib.mlir.dialects import mhlo

from .jax_compat import token_type, register_effect


class MPIEffect:
    def __hash__(self):
        # enforce a constant (known) hash
        return hash("I love mpi4jax")


effect = register_effect(MPIEffect)


def default_primitive_impl(primitive):
    return functools.partial(xla.apply_primitive, primitive)


def as_mhlo_constant(val, dtype):
    if isinstance(val, mhlo.ConstantOp):
        return val

    return mhlo.ConstantOp(
        ir.DenseElementsAttr.get(
            _np.array([val], dtype=dtype), type=mlir.dtype_to_ir_type(_np.dtype(dtype))
        )
    ).result


def get_default_layouts(operands, order="c"):
    (token,) = token_type()
    layouts = []

    if order == "c":
        default_layout = lambda t: tuple(range(len(t.shape) - 1, -1, -1))
    elif order == "f":
        default_layout = lambda t: tuple(range(len(t.shape)))
    else:
        raise ValueError("Unknown order: {}".format(order))

    for op in operands:
        if isinstance(op, (ir.Value)):
            if op.type == token:
                layouts.append(())
            else:
                tensor_type = ir.RankedTensorType(op.type)
                layouts.append(default_layout(tensor_type))

        elif isinstance(op, ir.RankedTensorType):
            layouts.append(default_layout(op))

        elif op == token:
            layouts.append(())

        else:
            raise ValueError("Unknown operand type: {}".format(type(op)))

    return layouts


def to_mpi_handle(mpi_obj):
    """
    Returns the handle of the underlying C mpi object.

    Only defined for some MPI types (such as MPI_Comm), throws NotImplementedError
    otherwise.

    Note: This is not a pointer, but the actual C integer representation of the object.
    """
    return _np.uintp(_MPI._handleof(mpi_obj))


def to_mpi_ptr(mpi_obj):
    """
    Returns a pointer to the underlying C MPI object
    """
    return _np.uintp(_MPI._addressof(mpi_obj))


# store type names as strings since we cannot expect all type objects to be present on all platforms
MPI_TYPE_MAP = {
    "float32": "FLOAT",
    "float64": "DOUBLE",
    "float128": "LONG_DOUBLE",
    "complex64": "COMPLEX",
    "complex128": "DOUBLE_COMPLEX",
    "int8": "INT8_T",
    "int16": "INT16_T",
    "int32": "INT32_T",
    "int64": "INT64_T",
    "uint8": "UINT8_T",
    "uint16": "UINT16_T",
    "uint32": "UINT32_T",
    "uint64": "UINT64_T",
    "bool": "BOOL",
}


def to_dtype_handle(dtype):
    """
    Returns the pointer to the MPI dtype of the input numpy dtype
    """
    dtype_name = _np.dtype(dtype).name
    if dtype_name not in MPI_TYPE_MAP:
        raise RuntimeError("Unknown MPI type for dtype {}".format(dtype_name))

    mpi_type = getattr(_MPI, MPI_TYPE_MAP[dtype_name])
    return to_mpi_handle(mpi_type)


# Helpers to make MPI objects hashable


class HashableMPIType:
    def __init__(self, obj):
        self.wrapped = obj

    def __hash__(self):
        return int(to_mpi_ptr(self.wrapped))


def wrap_as_hashable(obj):
    if isinstance(obj, HashableMPIType):
        return obj

    return HashableMPIType(obj)


def unpack_hashable(obj):
    if isinstance(obj, HashableMPIType):
        return obj.wrapped

    return obj


# Miscellaneous Utilities


def has_cuda_support() -> bool:
    """Returns True if mpi4jax is built with CUDA support and can be used with GPU-based
    jax-arrays, False otherwise.
    """
    from . import xla_bridge

    return xla_bridge.HAS_GPU_EXT



_default_comm = None


def get_default_comm():
    global _default_comm
    if _default_comm is None:
        from mpi4py import MPI

        _default_comm = MPI.COMM_WORLD.Clone()

    return _default_comm



import numpy as _np
from mpi4py import MPI as _MPI

from jax import core
from jax.core import Primitive, Tracer, Token
from jax.interpreters import batching
from jax.lax import create_token

from jax.interpreters import mlir

from ..utils import (
    HashableMPIType,
    default_primitive_impl,
    to_mpi_handle,
    unpack_hashable,
    wrap_as_hashable,
    as_mhlo_constant,
    get_default_layouts,
    effect,
)
from ..jax_compat import hlo_custom_call, token_type
from ..decorators import translation_rule_cpu, translation_rule_gpu
from ..validation import enforce_types
from ..comm import get_default_comm


# The Jax primitive
mpi_barrier_p = Primitive("barrier_mpi")  # Create the primitive
mpi_barrier_impl = default_primitive_impl(mpi_barrier_p)


# This function applies the primitive to an AST
@enforce_types(
    comm=(type(None), _MPI.Intracomm, HashableMPIType),
    token=(type(None), Token, Tracer),
)
def barrier(*, comm=None, token=None):
    """Perform a barrier operation.

    Arguments:
        comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
            a clone of :obj:`COMM_WORLD`).
        token (Token): XLA token to use to ensure correct execution order.
            If not given, a new token is generated.

    Returns:
        Token:
            - A new, modified token, that depends on this operation.

    """
    if token is None:
        token = create_token()

    if comm is None:
        comm = get_default_comm()

    comm = wrap_as_hashable(comm)
    return mpi_barrier_p.bind(token, comm=comm)


# This function compiles the operation
# transpose is a boolean flag that signals whever this is the forward pass
# performing the MPI reduction, or the transposed pass, which is trivial
@translation_rule_cpu
def mpi_barrier_xla_encode_cpu(ctx, token, comm):
    comm = unpack_hashable(comm)

    out_types = token_type()

    operands = (
        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
        token,
    )

    # JAX insists on outputs being iterable
    return [
        hlo_custom_call(
            b"mpi_barrier",
            out_types=out_types,
            operands=operands,
            operand_layouts=get_default_layouts(operands),
            result_layouts=get_default_layouts(out_types),
            has_side_effect=True,
        )
    ]


@translation_rule_gpu
def mpi_barrier_xla_encode_gpu(ctx, token, comm):
    from ..xla_bridge.mpi_xla_bridge_gpu import build_barrier_descriptor

    comm = unpack_hashable(comm)

    out_types = token_type()

    operands = (token,)

    descriptor = build_barrier_descriptor(to_mpi_handle(comm))

    # JAX insists on outputs being iterable
    return [
        hlo_custom_call(
            b"mpi_barrier",
            out_types=out_types,
            operands=operands,
            operand_layouts=get_default_layouts(operands),
            result_layouts=get_default_layouts(out_types),
            has_side_effect=True,
            backend_config=descriptor,
        )
    ]


# This function evaluates only the shapes during AST construction
def mpi_barrier_abstract_eval(token, comm):
    return core.abstract_token, {effect}


def mpi_barrier_batch_eval(in_args, batch_axes, comm):
    token = in_args[0]
    res = mpi_barrier_p.bind(token, comm=comm)
    return res, batch_axes


mpi_barrier_p.def_impl(mpi_barrier_impl)
mpi_barrier_p.def_effectful_abstract_eval(mpi_barrier_abstract_eval)

batching.primitive_batchers[mpi_barrier_p] = mpi_barrier_batch_eval

# assign to the primitive the correct encoder
mlir.register_lowering(mpi_barrier_p, mpi_barrier_xla_encode_cpu, platform="cpu")
mlir.register_lowering(mpi_barrier_p, mpi_barrier_xla_encode_gpu, platform="cuda")



import pytest

import warnings
import importlib


TEST_VERSIONS = {
    "0.0.0": (0, 0, 0),
    "0.1.61": (0, 1, 61),
    "0.1.61+cuda110": (0, 1, 61),
    "0.1.61a1": (0, 1, 61),
    "0.0.0.0": (0, 0, 0),
    "0.1.60.dev": (0, 1, 60),
}


@pytest.mark.parametrize(
    "fake_version",
    list(TEST_VERSIONS.keys()),
)
def test_versiontuple(fake_version):
    from mpi4jax._src.jax_compat import versiontuple

    assert versiontuple(fake_version) == TEST_VERSIONS[fake_version]


def test_version_warning(monkeypatch):
    import mpi4jax
    import jax

    # raise warning on too recent jax
    with monkeypatch.context() as m:
        m.setattr(jax, "__version__", "99.99.99")
        m.setenv("MPI4JAX_NO_WARN_JAX_VERSION", "0")
        with pytest.warns(UserWarning) as w:
            importlib.reload(mpi4jax._src)
            assert "but you have 99.99.99" in str(w[0])

    # do not raise warning on outdated jax
    with monkeypatch.context() as m:
        m.setattr(jax, "__version__", "00.00.00")
        m.setenv("MPI4JAX_NO_WARN_JAX_VERSION", "0")
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            importlib.reload(mpi4jax._src)

    # do not raise when envvar is set
    with monkeypatch.context() as m:
        m.setattr(jax, "__version__", "99.99.99")
        m.setenv("MPI4JAX_NO_WARN_JAX_VERSION", "1")
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            importlib.reload(mpi4jax._src)



import os
import sys
import shlex

from setuptools import setup, find_packages
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext

here = os.path.abspath(os.path.dirname(__file__))

# ensure vendored versioneer is on path
sys.path.insert(0, here)
import versioneer  # noqa: E402

try:
    from Cython.Build import cythonize
except ImportError:
    HAS_CYTHON = False
else:
    HAS_CYTHON = True

try:
    import mpi4py
except ImportError:
    HAS_MPI4PY = False
else:
    HAS_MPI4PY = True


##############
# Requirements

JAX_MINIMUM_VERSION = "0.3.25"

BASE_DEPENDENCIES = ["mpi4py>=3.0.1", "numpy", f"jax>={JAX_MINIMUM_VERSION}"]

DEV_DEPENDENCIES = [
    "pytest>=6",
    "pytest-cov>=2.10.1",
    "coverage[toml]>=5",
    "pre-commit",
    "black==21.6b0",
    "flake8==3.9.2",
    "tqdm>=4.52",
]

CYTHON_SUBMODULE_NAME = "mpi4jax._src.xla_bridge"
CYTHON_SUBMODULE_PATH = "mpi4jax/_src/xla_bridge"


#######
# Utils
def search_on_path(filenames):
    path_dirs = os.getenv("PATH", "").split(os.pathsep)

    for p in path_dirs:
        for filename in filenames:
            full_path = os.path.join(p, filename)
            if os.path.exists(full_path):
                return os.path.abspath(full_path)

    return None


def print_warning(*lines):
    print("**************************************************")
    for line in lines:
        print("*** WARNING: %s" % line)
    print("**************************************************")


# /end Utils
############


class custom_build_ext(build_ext):
    def build_extensions(self):
        config = mpi4py.get_config()
        mpi_compiler = os.environ.get("MPI4JAX_BUILD_MPICC", config["mpicc"])
        mpi_cmd = shlex.split(mpi_compiler)

        for exe in ("compiler", "compiler_so", "compiler_cxx", "linker_so"):
            # peel off compiler executable but keep flags
            current_flags = getattr(self.compiler, exe)[1:]
            self.compiler.set_executable(exe, [*mpi_cmd, *current_flags])

        build_ext.build_extensions(self)


################
# Cuda detection

# Taken from CUPY (MIT License)
def get_cuda_path():
    nvcc_path = search_on_path(("nvcc", "nvcc.exe"))
    cuda_path_default = None
    if nvcc_path is not None:
        cuda_path_default = os.path.normpath(
            os.path.join(os.path.dirname(nvcc_path), "..")
        )

    cuda_path = os.getenv("CUDA_PATH", "")  # Nvidia default on Windows
    if len(cuda_path) == 0:
        cuda_path = os.getenv("CUDA_ROOT", "")  # Nvidia default on Windows

    if len(cuda_path) > 0 and cuda_path != cuda_path_default:
        print_warning(
            "nvcc path != CUDA_PATH",
            "nvcc path: %s" % cuda_path_default,
            "CUDA_PATH: %s" % cuda_path,
        )

    if os.path.exists(cuda_path):
        _cuda_path = cuda_path
    elif cuda_path_default is not None:
        _cuda_path = cuda_path_default
    elif os.path.exists("/usr/local/cuda"):
        _cuda_path = "/usr/local/cuda"
    else:
        _cuda_path = None

    return _cuda_path


def get_cuda_info():
    cuda_info = {"compile": [], "libdirs": [], "libs": []}
    cuda_path = get_cuda_path()
    if not cuda_path:
        return cuda_info

    incdir = os.path.join(cuda_path, "include")
    if os.path.isdir(incdir):
        cuda_info["compile"].append(incdir)

    for libdir in ("lib64", "lib"):
        full_dir = os.path.join(cuda_path, libdir)
        if os.path.isdir(full_dir):
            cuda_info["libdirs"].append(full_dir)

    cuda_info["libs"].append("cudart")
    return cuda_info


cuda_info = get_cuda_info()

# /end Cuda detection
#####################


def get_extensions():
    cmd = sys.argv[1]
    require_extensions = any(
        cmd.startswith(subcmd) for subcmd in ("install", "build", "bdist", "develop")
    )

    if not HAS_MPI4PY or not HAS_CYTHON:
        # this should only happen when using python setup.py
        # or pip install --no-build-isolation
        if require_extensions:
            raise RuntimeError("Building mpi4jax requires Cython and mpi4py")
        else:
            return []

    extensions = [
        Extension(
            name=f"{CYTHON_SUBMODULE_NAME}.{mod}",
            sources=[f"{CYTHON_SUBMODULE_PATH}/{mod}.pyx"],
        )
        for mod in ("mpi_xla_bridge", "mpi_xla_bridge_cpu")
    ]

    if cuda_info["compile"] and cuda_info["libdirs"]:
        extensions.append(
            Extension(
                name=f"{CYTHON_SUBMODULE_NAME}.mpi_xla_bridge_gpu",
                sources=[f"{CYTHON_SUBMODULE_PATH}/mpi_xla_bridge_gpu.pyx"],
                include_dirs=cuda_info["compile"],
                library_dirs=cuda_info["libdirs"],
                libraries=cuda_info["libs"],
            )
        )
    else:
        print_warning("CUDA path not found", "(GPU extensions will not be built)")

    if HAS_CYTHON:
        extensions = cythonize(
            extensions,
            language_level=3,
        )

    return extensions


# get the long description from the README file
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


cmdclass = versioneer.get_cmdclass()
cmdclass.update(build_ext=custom_build_ext)

setup(
    name="mpi4jax",
    author="Filippo Vicentini",
    author_email="filippovicentini@gmail.com",
    description=(
        "Zero-copy MPI communication of JAX arrays, "
        "for turbo-charged HPC applications in Python ⚡"
    ),
    long_description=long_description,
    url="https://github.com/mpi4jax/mpi4jax",
    license="MIT",
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "Development Status :: 3 - Alpha",
        "Environment :: GPU :: NVIDIA CUDA",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering",
    ],
    packages=find_packages(),
    ext_modules=get_extensions(),
    python_requires=">=3.7",
    install_requires=BASE_DEPENDENCIES,
    extras_require={
        "dev": DEV_DEPENDENCIES,
    },
    package_data={"mpi4jax": ["_src/_latest_jax_version.txt"]},
    zip_safe=False,
)



import pytest

from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_alltoall():
    from mpi4jax import alltoall

    arr = jnp.ones((size, 3, 2)) * rank

    res, _ = alltoall(arr)
    for p in range(size):
        assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)


def test_alltoall_jit():
    from mpi4jax import alltoall

    arr = jnp.ones((size, 3, 2)) * rank

    res = jax.jit(lambda x: alltoall(x)[0])(arr)
    for p in range(size):
        assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)


def test_alltoall_wrong_size():
    from mpi4jax import alltoall

    arr = jnp.ones((size + 1,)) * rank
    with pytest.raises(ValueError) as excinfo:
        alltoall(arr)

    assert "must have shape (nproc, ...)" in str(excinfo.value)


def test_alltoall_transpose():
    # test whether consecutive transposes are behaving correctly under JIT
    # (some backends can optimize them out which leads to non-c-contiguous arrays)
    # see mpi4jax#176
    from mpi4jax import alltoall

    shape = (256, 256)

    def transpose(arr):
        arr = arr.reshape([shape[0] // size, size, shape[1] // size])
        arr = arr.transpose([1, 0, 2])
        arr, token = alltoall(arr, comm=comm)
        arr = arr.reshape([shape[0], shape[1] // size])
        arr = arr.transpose([1, 0])
        return arr, token

    transpose_jit = jax.jit(transpose)

    master_key = jax.random.PRNGKey(42)
    key = jax.random.split(master_key, size)[rank]
    arr = jax.random.normal(key, [shape[0] // size, shape[1]])

    assert jnp.array_equal(transpose(arr)[0], transpose_jit(arr)[0])



from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_gather():
    from mpi4jax import gather

    arr = jnp.ones((3, 2)) * rank

    res, _ = gather(arr, root=0)
    if rank == 0:
        for p in range(size):
            assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)
    else:
        assert jnp.array_equal(res, arr)


def test_gather_jit():
    from mpi4jax import gather

    arr = jnp.ones((3, 2)) * rank

    res = jax.jit(lambda x: gather(x, root=0)[0])(arr)
    if rank == 0:
        for p in range(size):
            assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)
    else:
        assert jnp.array_equal(res, arr)


def test_gather_scalar():
    from mpi4jax import gather

    arr = rank
    res, _ = gather(arr, root=0)
    if rank == 0:
        assert jnp.array_equal(res, jnp.arange(size))
    else:
        assert jnp.array_equal(res, arr)


def test_gather_scalar_jit():
    from mpi4jax import gather

    arr = rank
    res = jax.jit(lambda x: gather(x, root=0)[0])(arr)
    if rank == 0:
        assert jnp.array_equal(res, jnp.arange(size))
    else:
        assert jnp.array_equal(res, arr)
